# Standard library imports
from re import compile
from typing import TYPE_CHECKING

# Third party imports
from gspread import IncorrectCellLabel

if TYPE_CHECKING:
  type IntOrInf = int | float


R1C1_ADDR_ROW_COL_RE = compile(r"([rR](?P<rownum>[1-9]\d*))?([cC](?P<colnum>[1-9]\d*))?$")


def _r1c1_to_rowcol_unbounded(label: str) -> tuple[IntOrInf, IntOrInf]:
  """Translates a cell's address in R1C1 notation to a tuple of integers.

  :returns: a tuple containing `row` and `column` numbers. Both indexed
      from 1 (one).
  :rtype: tuple
  """
  match = R1C1_ADDR_ROW_COL_RE.match(label)
  if not match:
    raise IncorrectCellLabel(label)

  col, row = match.group("colnum", "rownum")

  col = int(col) if col is not None else float("inf")
  row = int(row) if row is not None else float("inf")
  return (row, col)


def r1c1_range_to_grid_range(name: str, sheet_id: int | None = None) -> dict[str, int]:
  """Converts a range defined in R1C1 notation to a dict representing
  a `GridRange`_.

  All indexes are zero-based. Indexes are half open, e.g the start
  index is inclusive and the end index is exclusive: [startIndex, endIndex).

  Missing indexes indicate the range is unbounded on that side.

  .. _GridRange: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#GridRange

  """
  start_label, _, end_label = name.partition(":")

  start_row_index, start_column_index = _r1c1_to_rowcol_unbounded(start_label)

  end_row_index, end_column_index = _r1c1_to_rowcol_unbounded(end_label or start_label)

  if start_row_index > end_row_index:
    start_row_index, end_row_index = end_row_index, start_row_index

  if start_column_index > end_column_index:
    start_column_index, end_column_index = end_column_index, start_column_index

  grid_range = {
    "startRowIndex": start_row_index - 1,
    "endRowIndex": end_row_index,
    "startColumnIndex": start_column_index - 1,
    "endColumnIndex": end_column_index,
  }

  filtered_grid_range: dict[str, int] = {key: value for (key, value) in grid_range.items() if isinstance(value, int)}

  if sheet_id is not None:
    filtered_grid_range["sheetId"] = sheet_id

  return filtered_grid_range



# Standard library imports
from logging import getLogger
from typing import TYPE_CHECKING, ClassVar

# First party imports
from aeth_ext.types import StrEnum

# Local folder imports
from .enums import LogActionEnum, SuppliersEnum

if TYPE_CHECKING:
  # Standard library imports
  from typing import Self

  # Local folder imports
  from . import CustomerID, InvoiceNum, StoreNum

logger = getLogger(__name__)


class ColNameEnum(StrEnum):
  __exclude__: ClassVar[list[str]] = []
  __init_include__: ClassVar[list[str]] = []
  __index_items__: ClassVar[list[str]] = []

  @classmethod
  def ordered_column_names(cls, *columns: str) -> list[str]:
    columns_list = [str(column) for column in columns]
    return [str(column) for column in cls if str(column) in columns_list]

  @classmethod
  def all_columns(cls) -> list[str]:
    return [str(column) for column in cls if str(column) not in cls.__exclude__ and not str(column).startswith("_")]

  @classmethod
  def err_reporting_columns(cls) -> list[str]:
    """
    Return all columns that are not excluded and do not start with an underscore.
    This is used for error reporting.
    """
    return [
      "err_field_name",
      "err_reason",
      *[str(column) for column in cls if str(column) not in cls.__exclude__ and not str(column).startswith("_")],
    ]

  @classmethod
  def init_columns(cls) -> list[str]:
    if not cls.__init_include__:
      return cls.all_columns()
    return [str(column) for column in cls if str(column) in cls.__init_include__ and not str(column).startswith("_")]

  @classmethod
  def testing_columns(cls) -> list[str]:
    return [str(column) for column in cls if str(column) not in cls.__exclude__]

  @classmethod
  def true_all_columns(cls) -> list[str]:
    return [str(column) for column in cls]

  @classmethod
  def get_enum_index(cls, value: Self) -> int:
    return list(cls).index(value)


class DatabaseScheduleColumns(ColNameEnum):
  __index_items__: ClassVar[list[str]] = ["supplier", "store"]

  supplier = "supplier"
  store = "store"
  customer = "customer"
  state = "state"
  expected_delivery_day = "expected_delivery_day"
  invoice_pickup_time = "invoice_pickup_time"
  invoice_dropoff_time = "invoice_dropoff_time"
  invoice_grabbed = "invoice_grabbed"
  invoice_applied = "invoice_applied"
  manually_moved = "manually_moved"


type DatabaseScheduleIndex = tuple[SuppliersEnum, StoreNum]


class DatabaseOrderLogColumns(ColNameEnum):
  __index_items__: ClassVar[list[str]] = ["supplier", "store", "invoice_number", "customer", "action", "status", "action_datetime"]

  supplier = "supplier"
  store = "store"
  invoice_number = "invoice_number"
  customer = "customer"
  action = "action"
  status = "status"
  action_datetime = "action_datetime"
  week_end_date = "week_end_date"
  notes = "notes"


type DatabaseOrderLogIndex = (
  tuple[SuppliersEnum, StoreNum, InvoiceNum, CustomerID, LogActionEnum]
  | tuple[SuppliersEnum, StoreNum, InvoiceNum, slice, LogActionEnum]
  | tuple[SuppliersEnum, slice, InvoiceNum, CustomerID, LogActionEnum]
)
